//   Copyright (c)  2006  John Abbott
//   Main authors: Laura Torrente (assisted by John Abbott)

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/ApproxPts.H"
#include "CoCoA/assert.H"

#include <iostream>
using std::ostream;


//#include <vector>
using std::vector;
#include <list>
using std::list;
#include <algorithm>
using std::copy;
using std::find_if;
#include <cmath>
using std::log; // only in UnitVolBallRadius
using std::exp; // only in UnitVolBallRadius

namespace CoCoA
{

  namespace ApproxPts
  {

    namespace // anonymous namespace for file local auxiliary functions and definitions.
    {

      // This is not really specific to approximate points, but where else should it go?
      // Compute radius of D-ball having volume 1.
      // (I got the formula from Wikipedia under "hypersphere")
      double UnitVolBallRadius(size_t D)
      {
        CoCoA_ASSERT(D < 10000); // disallow very high dimensions
        const double pi = 4*atan(1.0);
        double logfactor = (D/2)*(-log(2*pi)); // D/2 exploits integer division!  NB -(D/2)*log(...) does NOT WORK.
        if (D & 1)
          logfactor -= log(2.0);
        for (size_t k=D; k > 1; k -= 2)
          logfactor += log(double(k));
        return std::exp(logfactor/D);
      }


      // Compute square of 2-norm between P and Q.
      double L2NormSq(const ApproxPt& P, const ApproxPt& Q)
      {
        CoCoA_ASSERT(P.size() == Q.size());
        double SumSq = 0;
        const size_t dim = P.size();
        for (size_t i=0; i < dim; ++i)
          SumSq += square(P[i]-Q[i]);
        return SumSq;
      }


      // Data structure for representing PreProcessed Points.
      struct PPP
      {
        ApproxPt myCoords;      // the coords of the preprocessed point
        size_t myNumOrigPts;    // length of list myOrigPts (used only in PreprocessSubdivAlgm)
        list<size_t> myOrigPts; // indices into a master table
        list<double> myDistanceSq;  // used only in PreprocessSubdivAlgm
        double mySumDistanceSq;     // used only in PreprocessSubdivAlgm
      };


      // Returns the average of a set of ApproxPts; the set is indicated by the first arg
      // which is a list of indices into the second arg.
      ApproxPt ComputeCofG(const list<size_t>& IndexList, const vector<ApproxPt>& OrigPts)
      {
        const size_t dim = OrigPts[0].size();
        ApproxPt CofG(dim);
        for (list<size_t>::const_iterator it=IndexList.begin(); it != IndexList.end(); ++it)
        {
          const ApproxPt& P = OrigPts[*it];
          for (size_t i=0; i < dim; ++i)
          {
            CofG[i] += P[i];
          }
        }
        const size_t n = IndexList.size();
        for (size_t i=0; i < dim; ++i)
          CofG[i] /= n;
        return CofG;
      }


      // Returns true if the values are sane; false otherwise.
      bool CheckSemiWidths(const vector<double>& semiwidth)
      {
        const size_t NumDims = semiwidth.size();
        for (size_t i=0; i < NumDims; ++i)
          if (semiwidth[i] <= 0) return false;
        return true;
      }


      // Rescale coords of pt by factors in array ScaleFactor.
      void ScalePt(ApproxPt& pt, const vector<double>& ScaleFactor)
      {
        const size_t dim = ScaleFactor.size();
        for (size_t i=0; i < dim; ++i)
        {
          pt[i] *= ScaleFactor[i];
        }
      }

      // Rescale in place the OrigPts.
      void NormalizeOrigPts(vector<ApproxPt>& OrigPts, const vector<double>& semiwidth)
      {
        const size_t NumPts = OrigPts.size();
        const size_t dim = semiwidth.size();

        vector<double> InverseSemiwidth(dim);
        for (size_t j=0; j < dim; ++j)
          InverseSemiwidth[j] = 1/semiwidth[j];

        for (size_t i=0; i < NumPts; ++i)
          ScalePt(OrigPts[i], InverseSemiwidth);
      }


      /////////////////////////////////////////////////////////////////////////////
      // Auxiliary definitions for the grid algorithm.

      ApproxPt GridNearPoint(const ApproxPt& P)
      {
        const size_t dim = P.size();
        ApproxPt GNP(dim);
        for (size_t i=0; i < dim; ++i)
          GNP[i] = 2*round(P[i]/2);
        return GNP;
      }


      // Only used in GridAlgm as arg to std::find.
      // WARNING data member is a reference!!
      class CoordsEq
      {
      public:
        CoordsEq(const ApproxPt& P): myApproxPt(P) {};
        bool operator()(const PPP& PPPt) { return myApproxPt == PPPt.myCoords; };
      private: // data members
        const ApproxPt& myApproxPt;
      };


      /////////////////////////////////////////////////////////////////////////////
      // Functions for aggregative algorithm

      // ave = (w1*P1 + w2*P2)/(w1+w2);
      void WeightedAve(ApproxPt& ave, double w1, const ApproxPt& P1, double w2, const ApproxPt& P2)
      {
        const size_t dim = ave.size();
        CoCoA_ASSERT(P1.size() == dim && P2.size() == dim);
        CoCoA_ASSERT(w1+w2 != 0);
        for (size_t i=0; i < dim; ++i)
        {
          ave[i] = (w1*P1[i]+w2*P2[i])/(w1+w2);
        }
      }


      struct NearPair
      {
      public:
        NearPair(size_t i, size_t j, double sepsq): myLowerIndex(i), myHigherIndex(j), mySepSq(sepsq) {};
      public: // data members
        const size_t myLowerIndex, myHigherIndex;
        double mySepSq;
      };


      // Used only as arg to std::sort algorithm
      inline bool LessNearPair(const NearPair& P1, const NearPair& P2)
      {
        return P1.mySepSq < P2.mySepSq;
      }


      // Used in call to algorithm std::remove_if.
      class NearPairContains
      {
      public:
        NearPairContains(size_t j): myIndex(j) {}
        bool operator()(const NearPair& P) { return P.myLowerIndex == myIndex || P.myHigherIndex == myIndex; }
      private: // data member
        const size_t myIndex;
      };




      /////////////////////////////////////////////////////////////////////////////
      // Below is code for the "subdiv" algorithm.


      // Fill in data members myDistanceSq and mySumDistanceSq of X
      void ComputeDistSq(PPP& X, const vector<ApproxPt>& OrigPts)
      {
        X.myDistanceSq.clear();
        double SumDistanceSq = 0;
        const list<size_t>::const_iterator finished = X.myOrigPts.end();
        for (list<size_t>::const_iterator it = X.myOrigPts.begin(); it != finished; ++it)
        {
          double DistanceSq = L2NormSq(X.myCoords, OrigPts[*it]);
          X.myDistanceSq.push_back(DistanceSq);
          SumDistanceSq += DistanceSq;
        }
        X.mySumDistanceSq = SumDistanceSq;
      }


      // Make Y a copy of X but without the i-th OrigPt.
      void EraseAndUpdate(PPP& Y, const PPP& X, size_t i, const vector<ApproxPt>& V)
      {
        Y.myOrigPts = X.myOrigPts;
        Y.myOrigPts.remove(i);
        Y.myNumOrigPts = X.myNumOrigPts-1;
        Y.myCoords = ComputeCofG(Y.myOrigPts, V);
        ComputeDistSq(Y, V);
      }


      // Make Y a copy of X with the i-th OrigPt added in.
      void AddAndUpdate(PPP& Y, const PPP& X, size_t i, const vector<ApproxPt>& V)
      {
        Y.myOrigPts = X.myOrigPts;
        Y.myOrigPts.push_back(i);
        Y.myNumOrigPts = X.myNumOrigPts+1;
        Y.myCoords = ComputeCofG(Y.myOrigPts, V);
        ComputeDistSq(Y, V);
      }


      // Redistribute the assignments of OrigPts to PPPts to minimize the sum of squares of distances
      void redistribute(vector<PPP>& PPPts, const vector<ApproxPt>& OrigPts)
      {
        const size_t NumPPP = PPPts.size();
        bool MadeAnImprovement = true;
        while (MadeAnImprovement)
        {
          MadeAnImprovement = false;
          double MaxGain = 0.000001; // morally zero but must allow for rounding error
          size_t index_from = 0; // index of PPPt we move from (initialize to keep compiler quiet)
          size_t index_to = 0;   // index of PPPt we move to (initialize to keep compiler quiet)
          size_t IndexOrigPt = 0; // index of orig point we want to move (initialize to keep compiler quiet)

          // For each PPPt i do ...
          for (size_t i=0; i < NumPPP; ++i)
          {
            const double Ni = double(PPPts[i].myNumOrigPts);
            if (Ni == 1) continue;
            const double coeffi = Ni/(Ni-1);
            // For each index (of OrigPt associated to PPPt i) it do...
            list<double>::iterator it_dist = PPPts[i].myDistanceSq.begin();
            for (list<size_t>::iterator it = PPPts[i].myOrigPts.begin(); it != PPPts[i].myOrigPts.end(); ++it, ++it_dist)
            {
              const double diffi = (*it_dist)*coeffi;

              // For each PPPt j (difft from PPPt i) do...
              for (size_t j=0; j < NumPPP; ++j)
              {
                if (j == i) continue;
                const double Nj = double(PPPts[j].myNumOrigPts);
                const double coeffj = Nj/(Nj+1);
                const double diffj = -coeffj * L2NormSq(PPPts[j].myCoords, OrigPts[*it]);
                const double gain = diffi + diffj;
                if (gain > MaxGain)
                {
                  MadeAnImprovement = true;
                  MaxGain = gain;
                  index_from = i;
                  index_to = j;
                  IndexOrigPt = *it;
                }
              }
            }
          }

          if (MadeAnImprovement)
          {
            // Move orig point to new PPPt...
            AddAndUpdate(PPPts[index_to], PPPts[index_to], IndexOrigPt, OrigPts);
            EraseAndUpdate(PPPts[index_from], PPPts[index_from], IndexOrigPt, OrigPts);
          }
        }
      }


      // Initialize the list of preprocessed points for subdivision algm.
      void PreprocessSubdivInit(vector<PPP>& PPPts, const vector<ApproxPt>& OrigPts)
      {
        CoCoA_ASSERT(PPPts.empty());
        PPPts.resize(1);
        const size_t NumPts = OrigPts.size();
        for (size_t i=0; i < NumPts; ++i)
          PPPts[0].myOrigPts.push_back(i);
        PPPts[0].myNumOrigPts = NumPts;
        PPPts[0].myCoords = ComputeCofG(PPPts[0].myOrigPts, OrigPts);
        ComputeDistSq(PPPts[0], OrigPts);
      }


      // Determine if some OrigPt is too far from its corresponding PPPt.
      // If result is true, then WorstOrigIdx and WorstPPPIdx contain the indices
      // of the most distant pair of orig and preprocessed points.
      bool ExistsBadPoint(size_t& WorstOrigIdx, size_t& WorstPPPIdx, list<size_t>::iterator& it_list_worst, /*const*/ vector<PPP>& V)
      {
        it_list_worst = V[0].myOrigPts.begin();
        list<double>::iterator it_dist_worst = V[0].myDistanceSq.begin();
        const size_t NumPPP = V.size();
        WorstPPPIdx = 0;
        for (size_t i=0; i < NumPPP; ++i)
        {
          list<size_t>::iterator it_list = V[i].myOrigPts.begin();
          for (list<double>::iterator it_dist = V[i].myDistanceSq.begin(); it_dist != V[i].myDistanceSq.end(); ++it_dist)
          {
            if (*it_dist > *it_dist_worst)
            {
              it_dist_worst = it_dist;
              it_list_worst = it_list;
              WorstPPPIdx = i;
            }
            ++it_list;
          }
        }
        const size_t dim = V[0].myCoords.size();
        const double CriticalRadius = 2*UnitVolBallRadius(dim); // inefficient if dim is large and num of pts is small
        if (*it_dist_worst < CriticalRadius) return false;
        WorstOrigIdx = *it_list_worst;
        return true;
      }


      // Adjoin a new preprocessed point onto the end of V.
      // The new point comprises just one OrigPt, the one with index OrigPtIdx.
      void NewPPP(vector<PPP>& PPPts, const ApproxPt& P, size_t OrigPtIdx)
      {
        const size_t NumPPPts = PPPts.size();
        PPPts.resize(NumPPPts+1);
        PPP& NewPt(PPPts[NumPPPts]);
        NewPt.myCoords = P;
        NewPt.myOrigPts.push_back(OrigPtIdx);
        NewPt.myNumOrigPts = 1;
        NewPt.myDistanceSq.push_back(0);
        NewPt.mySumDistanceSq = 0;
      }


    } // end of anonymous namespace


    /////////////////////////////////////////////////////////////////////////////
    // Below are the three main algorithms.
    // First the "grid" algorithm.

    void PreprocessGridAlgm(vector<ApproxPt>& NewPts, vector<ApproxPt> OrigPts, const vector<double>& semiwidth)
    {
      const size_t NumPts = OrigPts.size();

      // Some sanity checks on the inputs.
      CoCoA_ASSERT(NumPts > 0);
      CoCoA_ASSERT(OrigPts[0].size() > 0);
      CoCoA_ASSERT(semiwidth.size() == OrigPts[0].size());
      CoCoA_ASSERT(CheckSemiWidths(semiwidth)); // give real error???

      NormalizeOrigPts(OrigPts, semiwidth);


      vector<PPP> PPPts; PPPts.reserve(NumPts);
      // For each OrigPts[i] do...
      for (size_t i = 0; i < NumPts; ++i)
      {
        const ApproxPt P = GridNearPoint(OrigPts[i]);
        vector<PPP>::iterator pos = find_if(PPPts.begin(), PPPts.end(), CoordsEq(P));
        if (pos != PPPts.end())
          pos->myOrigPts.push_back(i);         // adjoin to existing class
        else
          NewPPP(PPPts, P, i);  // add a new class
      }

      const size_t NumPPPts = PPPts.size();
      NewPts.resize(NumPPPts);
      for (size_t k = 0; k < NumPPPts; ++k)
      {
        NewPts[k] = ComputeCofG(PPPts[k].myOrigPts, OrigPts);
        ScalePt(NewPts[k], semiwidth);
      }

    }


    /////////////////////////////////////////////////////////////////////////////
    // The "aggregative" algorithm.

    void PreprocessAggrAlgm(vector<ApproxPt>& NewPts, vector<ApproxPt> OrigPts, const vector<double>& semiwidth)
    {
      const size_t NumPoints = OrigPts.size();
      const size_t dim = OrigPts[0].size();

      // Some sanity checks on the inputs.
      CoCoA_ASSERT(NumPoints > 0);
      CoCoA_ASSERT(dim > 0);
      CoCoA_ASSERT(semiwidth.size() == dim);
      CoCoA_ASSERT(CheckSemiWidths(semiwidth)); // give real error???

      // Normalization effectively makes the semiwidth in each direction equal to 1.
      NormalizeOrigPts(OrigPts, semiwidth);
      const double CriticalRadiusSq = square(2*UnitVolBallRadius(dim));

      vector<PPP> PreprocessedPoints(NumPoints);
      for (size_t i=0; i < NumPoints; ++i)
      {
        PreprocessedPoints[i].myCoords = OrigPts[i];
        PreprocessedPoints[i].myOrigPts.push_back(i);
      }

      // nbrs contains all pairs which could conceivably collapse together
      // (i.e. only pairs separated by at most the CriticalRadius)
      list<NearPair> nbrs;
      for (size_t i=0; i < NumPoints; ++i)
      {
        for (size_t j=i+1; j < NumPoints; ++j)
	{
	  const double d2 = L2NormSq(PreprocessedPoints[i].myCoords,
                                     PreprocessedPoints[j].myCoords);
	  if (d2 < 4*CriticalRadiusSq)
	    nbrs.push_back(NearPair(i,j,d2));

	}
      }

      nbrs.sort(LessNearPair);

      ApproxPt CofG(dim);
      while (!nbrs.empty() && nbrs.begin()->mySepSq < 4*CriticalRadiusSq)
      {
        const size_t i = nbrs.front().myLowerIndex;
        const size_t j = nbrs.front().myHigherIndex;
        //cout << "pair is " << i << "  " << j << endl;
        //cout << "dist=" << nbrs.front().mySepSq << endl;
        nbrs.erase(nbrs.begin()); // delete first elem of nbrs

        const PPP& Pi = PreprocessedPoints[i];
        const PPP& Pj = PreprocessedPoints[j];
        // Compute CofG of the union of Pi an Pj...
        WeightedAve(CofG, Pi.myOrigPts.size(), Pi.myCoords, Pj.myOrigPts.size(), Pj.myCoords);

        // Check that all original points in the union are close to CofG.
        bool OK = true;
        for (list<size_t>::const_iterator it=Pi.myOrigPts.begin(); OK && it != Pi.myOrigPts.end(); ++it)
	{
	  OK &= (L2NormSq(CofG, OrigPts[*it]) < CriticalRadiusSq);
	}
        for (list<size_t>::const_iterator it=Pj.myOrigPts.begin(); OK && it != Pj.myOrigPts.end(); ++it)
	{
	  OK &= (L2NormSq(CofG, OrigPts[*it]) < CriticalRadiusSq);
	}
        if (!OK) continue; // We cannot unite these two points, so go on to next pair of points.

        // The union is fine, so accept it.  We put all data into i-th PPP
        // and empty the data of the j-th PPP.
        swap(CofG, PreprocessedPoints[i].myCoords);
        {
          list<size_t>& PPPOrigPts = PreprocessedPoints[i].myOrigPts;
          PPPOrigPts.splice(PPPOrigPts.begin(), PreprocessedPoints[j].myOrigPts);
        }
        nbrs.remove_if(NearPairContains(j)); // since Pj has been eliminated.
        // Update all entries refering to i-th point.
        for (list<NearPair>::iterator nbr = nbrs.begin(); nbr != nbrs.end(); ++nbr)
	{
	  if (!NearPairContains(i)(*nbr)) continue;
	  size_t other = nbr->myLowerIndex;
	  if (other == i) other = nbr->myHigherIndex;
	  nbr->mySepSq = L2NormSq(CofG, PreprocessedPoints[other].myCoords);
	}
        nbrs.sort(LessNearPair); // put the list back into order -- SLUG SLUG SLUG!!!
        //cout << "len nbrs = " << nbrs.size() << endl;
      }

      // There are no futher possible mergings, so copy answer into NewPts.
      NewPts.clear();
      for (size_t i=0; i < NumPoints; ++i)
      {
        if (!PreprocessedPoints[i].myOrigPts.empty())
        {
          NewPts.push_back(PreprocessedPoints[i].myCoords);
          ScalePt(NewPts.back(), semiwidth);
        }
      }
    }


    /////////////////////////////////////////////////////////////////////////////
    // The "subdivision" algorithm.

    void PreprocessSubdivAlgm(vector<ApproxPt>& NewPts, vector<ApproxPt> OrigPts, const vector<double>& semiwidth)
    {
      const size_t NumPoints = OrigPts.size();

      // Some sanity checks on the inputs.
      CoCoA_ASSERT(NumPoints > 0);
      CoCoA_ASSERT(OrigPts[0].size() > 0);
      CoCoA_ASSERT(semiwidth.size() == OrigPts[0].size());
      CoCoA_ASSERT(CheckSemiWidths(semiwidth)); // give real error???

      // Normalization effectively makes the semiwidth in each direction equal to 1.
      NormalizeOrigPts(OrigPts, semiwidth);

      // Initially all orig points are associated to a single preprocessed point.
      vector<PPP> PPPoints; PPPoints.reserve(NumPoints);
      PreprocessSubdivInit(PPPoints, OrigPts);

      // Main loop.
      size_t WorstPPPIdx, WorstOrigIdx; // these are set in call to ExistsBadPoint
      list<size_t>::iterator it_worst;
      while (ExistsBadPoint(WorstOrigIdx, WorstPPPIdx, it_worst, PPPoints))
      {
        NewPPP(PPPoints, OrigPts[WorstOrigIdx], WorstOrigIdx);
        EraseAndUpdate(PPPoints[WorstPPPIdx], PPPoints[WorstPPPIdx], *it_worst, OrigPts);
        redistribute(PPPoints, OrigPts);
      }

      // Copy and denormalize result.
      NewPts.clear();
      NewPts.reserve(PPPoints.size());
      for (size_t i=0; i < PPPoints.size(); ++i)
      {
        NewPts.push_back(PPPoints[i].myCoords);
        ScalePt(NewPts[i], semiwidth);
      }

    }


  } // end of namespace ApproxPts

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/ApproxPts.C,v 1.2 2007/10/30 17:14:08 abbott Exp $
// $Log: ApproxPts.C,v $
// Revision 1.2  2007/10/30 17:14:08  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.8  2007/03/08 18:22:30  cocoa
// Just whitespace cleaning.
//
// Revision 1.7  2006/12/21 13:48:33  cocoa
// Made all increment/decrement calls prefix (except where the must be postfix).
//
// Revision 1.6  2006/11/22 14:43:32  cocoa
// -- minor cleaning (indicated by Intel compiler)
//
// Revision 1.5  2006/11/20 15:53:38  cocoa
// Minor cleaning.  Improved comments.
//
// Revision 1.4  2006/10/31 14:26:03  cocoa
// Added some missing consts; should make the code a little easier to comprehend.
//
// Revision 1.3  2006/10/06 14:04:15  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.2  2006/06/21 17:05:47  cocoa
// Major overhaul of approx point preprocessing algms.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.3  2006/05/29 16:16:47  cocoa
// Added "disj" preprocessing algorithm.
//
// Revision 1.2  2006/05/22 15:52:16  cocoa
// Added preprocess-disg algorithm to ApproxPts.
// Sundry minor improvements.
//
// Revision 1.1  2006/05/12 13:16:30  cocoa
// Added functions for preprocessing approximate points.
//
//
